/*
 * gc global variable provides access to GUI Composer infrastructure components and project information.
 * For more information, please see the Working with Javascript guide in the online help.
 */
var gc = gc || {};
gc.services = gc.services || {};

var rtls = rtls || {};
var scanResults = [];
var requestHandler = null;
const CONN_INT = 80 // 100ms connection interval (n*1.25)


const ws_send = (req) => {
    console.log('>> ' + JSON.stringify(req))
    gc.ws.send(JSON.stringify(req))
};

const log = (level, msg) => {
    console.log(msg);
    if (templateObj) templateObj.$.logview.log(level, msg)
}

const controlApi = {
    list_devices: () => ws_send({ control: { req: 'LIST_DEVICES' } })
};

const syncreq = (identifier, command, payload) => ({ identifier, message: { type: "SyncReq", subsystem: "RTLS", command, payload: payload || {} } })

const Device = (devInfo) => {
    const state = {};

    return {
        ...devInfo,
        isMaster: devInfo.caps.indexOf('RTLS_MASTER') > -1,
        identify: () => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_IDENTIFY')),
        reset: () => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_RESET_DEVICE')),
        scan: () => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_SCAN')),
        connect: (addrType, peerAddr, connInterval) => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_CONNECT', { addrType, peerAddr, connInterval})),
        terminate: () => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_TERMINATE_LINK')),
        setConnInfo: (accessAddress, connInterval, hopValue, mSCA, currChan, chanMap, crcInit) => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_CONN_PARAMS', { accessAddress, connInterval, hopValue, mSCA, currChan, chanMap, crcInit })),
        //
        calibrate: (enable, samplesPerFreq, calibDistance) => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_TOF_CALIBRATE', { enable, samplesPerFreq, calibDistance})),
        tofStart: (enable) => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_TOF_ENABLE', { enable })),
        tofSetParams: (tofRole, numSamples, numFreq, autoTofRssiThresh, resultMode, runMode, constSyncwords, frequencies) => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_TOF_SET_PARAMS', { tofRole, numSamples, numFreq, autoTofRssiThresh, resultMode, runMode, constSyncwords, frequencies })),
        tofSetSeed: (seed) => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_TOF_SET_SEC_SEED', { seed })),
        tofGetSeed: () => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_TOF_GET_SEC_SEED')),
        //
        aoaStart: (enable) => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_AOA_ENABLE', { enable })),
        aoaSetParams: (aoaRole, aoaResultMode, cteScanOvs, cteOffset, cteTime) => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_AOA_SET_PARAMS', { aoaRole, aoaResultMode, cteScanOvs, cteOffset, cteTime })),
        //
        state,
    }
}

const Request = (msg) => {
    const device = rtls.devices && rtls.devices.find(e => e.identifier === msg.identifier);
    return {
        device: device,
        ...msg.message,
    }
}


const handle_control = (control) => {
    switch (control.req) {
        case 'LIST_DEVICES':
            rtls.devices = control.devices.map(e => Device(e));

            rtls.master = rtls.devices.find(e => e.isMaster);
            rtls.passives = rtls.devices.filter(e => !e.isMaster);

            // Error check and mode check
            // 1. Need at least 1 master
            if (!rtls.master) log('error', 'Need a master device');

            // 2. All must be either AoA or ToF
            const allTof = rtls.devices.every(e => (e.caps.indexOf('TOF_MASTER') >= 0) || (e.caps.indexOf('TOF_PASSIVE') >= 0));
            const allNotTof = rtls.devices.every(e => !((e.caps.indexOf('TOF_MASTER') >= 0) || (e.caps.indexOf('TOF_PASSIVE') >= 0)));
            if (!(allTof || allNotTof)) log('error', 'All nodes must be either AoA or ToF');

            // 3. Need at least 1 passive for AoA
            if (allNotTof && rtls.passives.length === 0) log('error', 'Need at least 1 passive for AoA');

            // 4. Set up the command handler
            if (allTof) requestHandler = handle_request_tof;
            else requestHandler = handle_request_aoa;

            // Initialize graph
            if (templateObj) {
                const deviceLabels = control.devices.map(e => `${e.port} / ${e.identifier} / ${e.caps.join('+')}`).join('|');
                templateObj.$.lst_connected_devices.labels=deviceLabels;

                const graphNodes = allTof ? [rtls.master].concat(rtls.passives.slice(0,3)) : rtls.passives.slice(0,4);
                const graph = templateObj.$.scalargraph;
                graph.titleLabel = allTof ? 'Time of Flight' : 'Angle of Arrival';
                graph.fixedScale = true;
                graph.numChannels = graphNodes.length;
                graphNodes.forEach((e, i) => {
                    e.graphId = i;
                    graph['series_'+i+'_label'] = e.caps.join('+')
                    graph['series_'+i+'_units'] = allTof ? 'meters' : 'degrees';
                    graph['series_'+i+'_min'] = allTof ? 0 : -150;
                    graph['series_'+i+'_max'] = allTof ? 30 : 150;
                })
            }

            // Initialize devices
            scanResults = [];

            rtls.master.terminate();
            rtls.passives.forEach(it => it.reset());
            setTimeout(() => {
                rtls.master.reset();

                // Check if we are to automatically connect
                if(templateObj.$.chk_autortls.checked == true){
                    setTimeout(() => {
                        rtls.master.scan();
                    }, 500);
                }
                else{
                    templateObj.$.btn_scan.disabled = false;
                }
            }, 500);

            break;
        default:
            log('error', 'Unknown response: ' + control.req);
            break;
    }
}

let peerAddr;
let peerAddrType;
let seed;



const handle_request_aoa = (req) => {
    //
    // Connection establishment to remote device
    //

    // Save latest scan response
    if (req.command === 'RTLS_CMD_SCAN' && req.type === 'AsyncReq') {
        peerAddr = req.payload.addr;
        peerAddrType = req.payload.addrType;
        gui_scan_result(req);
    }

    // Connect to last found when scan is finished
    if (req.command === 'RTLS_CMD_SCAN_STOP') {
        if(templateObj.$.chk_autortls.checked == true){
            if(peerAddr){
                templateObj.$.ti_widget_progressbar.value = 25;
                templateObj.$.ti_widget_status_label.label = "Status: Connecting, Pairing, Opening L2CAP CoC";

                rtls.master.connect(peerAddrType, peerAddr, CONN_INT);
            }
            else {
                rtls.master.scan();
            }
        }
        else{
            templateObj.$.btn_conn_disconn.disabled = false;
        }
    }

    // When connected, let passives know about connection params
    if (req.command === 'RTLS_CMD_CONN_PARAMS' && req.type === 'AsyncReq') {
        const { accessAddress, connInterval, hopValue, mSCA, currChan, chanMap, crcInit } = req.payload;
        rtls.passives.forEach(e => {
            e.setConnInfo(accessAddress, connInterval, hopValue, mSCA, currChan+hopValue*1, chanMap, crcInit);
        });
    }

    //
    // Initialize AoA
    //

    // When master sends CONNECT later, pairing is complete and we can initialize AoA parameters
    if (req.command === 'RTLS_CMD_CONNECT' && req.type === 'AsyncReq' && req.payload.status === 'RTLS_SUCCESS' && req.device === rtls.master) {
        rtls.master.aoaSetParams('AOA_MASTER', 'AOA_MODE_ANGLE', 4, 4, 20);
        rtls.passives.forEach(it => {
           it.aoaSetParams('AOA_PASSIVE', 'AOA_MODE_ANGLE', 4, 4, 20);
        })

        templateObj.$.ti_widget_progressbar.value = 75;
        templateObj.$.ti_widget_status_label.label = "Status: Connected. Setting AoA Params";

        if(templateObj.$.chk_autortls.checked != true){
            templateObj.$.btn_enable_rtls.disabled = false;
        }
    }

    // After aoa parameters are set, start aoa
    if (req.command === 'RTLS_CMD_AOA_SET_PARAMS') {
        if(templateObj.$.chk_autortls.checked == true){
            templateObj.$.ti_widget_progressbar.value = 100;
            templateObj.$.ti_widget_status_label.label = "Status: Running. Sampling AoA data";
            req.device.aoaStart(1);
        }
    }

    //
    // Receive Data
    //

    if (req.command === 'RTLS_CMD_AOA_RESULT_ANGLE')
    {
        if (templateObj) {
            const graph = templateObj.$.scalargraph;
            const id = req.device.graphId;
            if (id !== undefined) {
                graph['value_'+id.toString()] = req.payload.angle;
            }
        }
    }
}

const handle_request_tof = (req) => {

    //
    // Initialize array for smoothing ToF distance results,
    // each device has its own array and count.
    // Averaging over 4 connection events gives us a reasonable confidence
    // level at +/- 1m confidence interval
    //
    var MAX_RES = 4
    req.device.state.dist_res = req.device.state.dist_res || [0.0, 0.0, 0.0, 0.0];
    req.device.state.conn_evt_count = req.device.state.conn_evt_count || 0;

    //
    // Local Variables related to ToF parameters and calibration
    //
    const freqList          = [2408, 2412, 2418, 2424];
    const numFreq           = freqList.length;
    const numSyncWords      = 256;
    const tofSampleMode     = 'TOF_MODE_DIST';
    const tofRunMode        = 'TOF_MODE_CONT';
    const tofRssiThreshold  = -55;
    const tofCalibDistance  = 1; // One meter
    const tofNumCalibSamples   = 1000;
    const constSyncwords = true;

    //
    // Connection establishment to remote device
    //

    // Save latest scan response
    if (req.command === 'RTLS_CMD_SCAN' && req.type === 'AsyncReq') {
        peerAddr = req.payload.addr;
        peerAddrType = req.payload.addrType;
        gui_scan_result(req);
    }

    // Connect to last found when scan is finished
    if (req.command === 'RTLS_CMD_SCAN_STOP') {

        if(templateObj.$.chk_autortls.checked == true){
            if(peerAddr){
                rtls.master.connect(peerAddrType, peerAddr, CONN_INT);

                templateObj.$.ti_widget_progressbar.value = 25;
                templateObj.$.ti_widget_status_label.label = "Status: Connecting, Pairing, Opening L2CAP CoC";
            }
            else {
                rtls.master.scan();
            }
        }
        else{
            templateObj.$.btn_conn_disconn.disabled = false;
        }
    }

    //
    // Initialize ToF
    //

    // When master sends CONNECT later, pairing is complete and we can initialize ToF parameters
    if (req.command === 'RTLS_CMD_CONNECT' && req.type === 'AsyncReq' &&
        req.payload.status === 'RTLS_SUCCESS') {

        var role = req.device === rtls.master ? 'TOF_MASTER' : 'TOF_PASSIVE'

        req.device.tofSetParams(role , numSyncWords, numFreq, tofRssiThreshold, tofSampleMode, tofRunMode,  constSyncwords, freqList);


        templateObj.$.ti_widget_progressbar.value = 50;
        templateObj.$.ti_widget_status_label.label = "Status: Connected. Setting ToF Params";

        // Reset object variables
        req.device.state.seedSet = false;
        req.device.state.calibSet = false;
        req.device.state.tofEnabled = false;
        req.device.state.calibDone = false;
        req.device.state.tofSet = false;
    }

    // After tof_init, get the random seed from the master
    if (req.command === 'RTLS_CMD_TOF_SET_PARAMS') {
        req.device.state.tofSet = true;

        if (rtls.passives.length > 0){
            // Make sure all nodes have parameters set before moving on
            if(rtls.passives.every(e => e.state.tofSet) &&
                rtls.master.state.tofSet === true){
                rtls.master.tofGetSeed();
            }
        }
        else{
            rtls.master.calibrate(true, tofNumCalibSamples, tofCalibDistance)
        }
    }

    // When we get seed, give it to the passives
    if (req.command === 'RTLS_CMD_TOF_GET_SEC_SEED') {
        seed = req.payload.seed;
        rtls.passives.forEach(e => {
            e.tofSetSeed(seed);
        })

        templateObj.$.ti_widget_progressbar.value = 70;
        templateObj.$.ti_widget_status_label.label = "Status: Connected. Distributing ToF security seed";
    }

    // When all seeds are set, run calibration
    if (req.command === 'RTLS_CMD_TOF_SET_SEC_SEED') {
        req.device.state.seedSet = true;

        if (rtls.passives.every(e => e.state.seedSet)) {
            rtls.passives.forEach(e => {
                e.calibrate(true, tofNumCalibSamples, tofCalibDistance);
            })

            rtls.master.calibrate(true, 0, tofCalibDistance)
            templateObj.$.ti_widget_progressbar.value = 70;
            templateObj.$.ti_widget_status_label.label = "Status: Connected, Calibrating ToF for distance mode.";
        }
    }

    // When calibration is enabled, enable ToF on passives
    if (req.command === 'RTLS_CMD_TOF_CALIBRATE' && req.type === 'SyncRsp') {
        req.device.state.calibSet = true;

        if(templateObj.$.chk_autortls.checked == true){
            // Enable ToF on passives after they have calibration enabled
            if(rtls.passives.length > 0){
                if(rtls.passives.every(e => e.state.calibSet) && req.device != rtls.master){
                    rtls.passives.forEach(e => {
                        e.tofStart(1);
                    })
                    templateObj.$.ti_widget_progressbar.value = 80;
                    templateObj.$.ti_widget_status_label.label = "Status: Connected, Calibrated. Enabling ToF on passive nodes.";
                }

            }
            else{
                rtls.master.tofStart(1);
                templateObj.$.ti_widget_progressbar.value = 100;
                templateObj.$.ti_widget_status_label.label = "Status: Status: Running. Sampling ToF data";
            }
        }
        else{
            templateObj.$.btn_enable_rtls.disabled = false;
        }
    }

    // Once all passives have ToF enabled, start ToF on master
    if (req.command === 'RTLS_CMD_TOF_ENABLE' && req.payload.status === 'RTLS_SUCCESS') {
        req.device.state.tofEnabled = true;

        // Once all passives have TOF started, start the master.
        if (rtls.passives.every(e => e.state.tofEnabled) && req.device != rtls.master)
        {
            rtls.master.tofStart(1);
        }
        templateObj.$.ti_widget_progressbar.value = 100;
        templateObj.$.ti_widget_status_label.label = "Status: Status: Running. Sampling ToF data";
    }

    if(req.command === 'RTLS_CMD_TOF_CALIBRATE' && req.type === 'AsyncReq' &&
       rtls.passives.includes(req.device))
    {
        req.device.state.calibDone = true;
        if (rtls.passives.every(e => e.state.calibDone))
        {
            rtls.master.calibrate(false, 0, tofCalibDistance)
        }
    }

    //
    // Receive Data
    //
    if (req.command === 'RTLS_CMD_TOF_RESULT_DIST')
    {
        // Add the new result to the end and shift out the previous
        req.device.state.dist_res[req.device.state.conn_evt_count++] = req.payload.distance

        if(req.device.state.conn_evt_count == MAX_RES)
        {
            // Reset the index
            req.device.state.conn_evt_count = 0;

            // Plot the average of the array
            if (templateObj) {
                const graph = templateObj.$.scalargraph;
                const id = req.device.graphId;
                if (id !== undefined) {
                    // Plot the average of the array at this time
                    graph['value_'+id.toString()] = req.device.state.dist_res.reduce((a,b) => a + b, 0) / req.device.state.dist_res.length;
                }
            }
        }
    }
}

//
// Function for handling received data from the websocket
//
const ws_receive = (msg) => {
    // console.log(msg.data);
    if (templateObj) templateObj.$.logview.log('info', msg.data);
    try {
        const parsed = JSON.parse(msg.data)

        if (parsed.control) {
            handle_control(parsed.control);
        } else if (rtls.devices) {
            const req = Request(parsed);
            if (req.command === 'RTLS_CMD_ERROR' && req.payload.status === 'RTLS_ILLEGAL_CMD') log('error', 'Illegal command received. Probably tried to do ToF on an AoA node or vice versa');
            // console.log(`${req.command} from ${req.device.name}`);

            if (!requestHandler) {
                log('error', 'No request handler, have not detected AoA or ToF nodes connected');
                return;
            }

            requestHandler(req);
        }
    } catch (err) {
        log('error', err);
    }
}

//
// OnClick handler for btn_scan. Tells master device to scan
// Only used when Auto RTLS is disabled, in case customer wants to choose
// slave based on BD_ADDR
//
const scan_for_slave = () => {
    // Disable the connect button while scanning
    templateObj.$.btn_conn_disconn.disabled = true;

    // Tell the master device to scan
    setTimeout(() => {
        rtls.master.scan();
    }, 500);

}

//
// OnClick handler for btn_conn_disconn. Tells master device to connect to slave
// Only used when Auto RTLS is disabled, in case customer wants to choose
// slave based on BD_ADDR. Takes the actively selected index of lst_remote_devices
//
const connect_disconnect = () => {
    // Get the index of the actively selected slave device
    var idx = templateObj.$.lst_remote_devices.getSelectedIndex();
    if(idx >= 0)
    {
        // If the index is valid, connect to that device
        rtls.master.connect(scanResults[idx].addrType, scanResults[idx].addr, CONN_INT);
        console.log("Connect to: " + scanResults[idx].addr);

        templateObj.$.ti_widget_progressbar.value = 25;
        templateObj.$.ti_widget_status_label.label = "Status: Connecting, Pairing, Opening L2CAP CoC";
    }
    else
    {
        console.warn("Can't connect, no item selected");
    }

}

//
// OnConnect handler for websocket. Initializes button states and lists devices
//
const wss_connect = () => {
    try {
        var ws_addr = templateObj.$.input_ws_addr.value
        gc.ws = new WebSocket(ws_addr);
        gc.ws.onmessage = ws_receive;
        gc.ws.onclose = () => {
            templateObj.$.logview.log('error', "Closed");
            templateObj.$.btn_connect.disabled = false;
        }
        gc.ws.onopen = () => {
            controlApi.list_devices();
            templateObj.$.ti_widget_progressbar.visible = true;
            templateObj.$.ti_widget_progressbar.value = 0;
            templateObj.$.ti_widget_status_label.visible = true;

            templateObj.$.ti_widget_status_label.label = "Status: Connected to Websocket";

            // Once we have connected to the websocket, we lock in the settings for Auto RTLS
            templateObj.$.btn_connect.disabled = true;
            templateObj.$.chk_autortls.disabled = true;
        }



    } catch (err) {
        templateObj.$.logview.log('error', "Could not connect")
        alert("Could not connect!");
        console.log(err);
    }
}

gc.tofs = gc.tofs || [];

//
// Handler for processing scan results and showing them in lst_remote_devices
//
const gui_scan_result = (req) => {
    if (!templateObj) return;
    const list = templateObj.$.lst_remote_devices;

    scanResults.push({...req.payload});
    list.labels = scanResults.map(e => `${e.addr}`).join('|');
}

const enable_rtls = (req) => {

    const allTof = rtls.devices.every(e => (e.caps.indexOf('TOF_MASTER') >= 0) || (e.caps.indexOf('TOF_PASSIVE') >= 0));
    if(allTof){
        if( rtls.passives.length > 0){
            if(rtls.passives.every(e => e.state.calibSet)){
                rtls.passives.forEach(e => {
                    e.tofStart(1);
                })
                templateObj.$.ti_widget_progressbar.value = 80;
                templateObj.$.ti_widget_status_label.label = "Status: Connected, Calibrated. Enabling ToF on passive nodes.";
            }
        }
        else{
            rtls.master.tofStart(1);
            templateObj.$.ti_widget_progressbar.value = 80;
            templateObj.$.ti_widget_status_label.label = "Status: Connected, Calibrated. Enabling ToF.";
        }
    }
    else{
        rtls.passives.forEach(e => {
            e.aoaStart(1);
        })
        rtls.master.aoaStart(1);

        templateObj.$.ti_widget_progressbar.value = 100;
        templateObj.$.ti_widget_status_label.label = "Status: Running. Sampling AoA data";
    }
}

/*
*  Boilerplate code for creating computed data bindings
*/
document.addEventListener('gc-databind-ready', function()
{
    /*
	*   Add custom computed value databindings here, using the following method:
	*
    *   function gc.databind.registry.bind(targetBinding, modelBinding, [getter], [setter]);
	*
	*   param targetBinding - single binding string or expression, or array of binding strings for multi-way binding.
	*   param modelBinding - single binding string or expression, or array of binding strings for multi-way binding.
	*   param getter - (optional) - custom getter function for computing the targetBinding value(s) based on modelBinding value(s).
	*   param setter - (optional) - custom setter function for computing the modelBinding value(s) based on targetBinding value(s).
    */

	// For example, a simple computed values based on simple expression
	// gc.databind.registry.bind('widget.id.propertyName', "targetVariable == 1 ? 'binding is one' : 'binding is not one'");

	// Or a custom two-way binding with custome getter and setter functions.  (setter is optional)  (getter only indicates one-way binding)
	// gc.databind.registry.bind('widget.id.propertyName', "targetVariable", function(value) { return value*5/9 + 32; }, function(value) { (return value-32)*9/5; });

	// Event 1 to n bindings
	/*
    gc.databind.registry.bind('widget.date.value',
        // dependant bindings needed in order to compute the date, in name/value pairs.
        {
            weekday: 'widget.dayOfWeek.selectedText',
            day: 'widget.dayOfMonth.value',
            month: 'widget.month.selectedText',
            year: 'widget.year.value'
        },
        // getter for date computation
        function(values)
        {
            // compute and return the string value to bind to the widget with id 'date'
            return values.weekday + ', ' + values.month + ' ' + values.day + ', ' + values.year;
        }
    );
	*/
});

/*
*  Boilerplate code for creating custom actions
*/
document.addEventListener('gc-nav-ready', function()
{
    /*
	*   Add custom actions for menu items using the following api:
	*
    *   function gc.nav.registryAction(id, runable, [isAvailable], [isVisible]);
	*
	*   param id - uniquely identifies the action, and should correspond to the action property of the menuaction widget.
	*   param runable - function that performs the custom action.
	*   param isAvailable - (optional) - function called when the menu action is about to appear.  Return false to disable the action, or true to enable it.
	*   param isVisible - (optional) - function called when the menu action is about to appear.  Return false to hide the action, or true to make it visible.
    */

	// For example,
	// gc.nav.registerAction('myCustomCloseAction', function() { window.close(); }, function() { return true; }, function() { return true; });

	// Alternatively, to programmatically disable a menu action at any time use:
	// gc.nav.disableAction('myCustomCloseAction);    then enable it again using:  gc.nav.enableAction('myCustomCloseAction');
});


/*
*  Boilerplate code for working with components in the application gist
*/


var initComplete = false;
var templateObj;

// Wait for DOMContentLoaded event before trying to access the application template
var init = function() {
    templateObj = document.querySelector('#template_obj');

    // Wait for the template to fire a dom-change event to indicate that it has been 'stamped'
    // before trying to access components in the application.
	templateObj.addEventListener('dom-change',function(){
	    if (initComplete) return;
	    this.async(function(){
    	    initComplete = true;
    	    console.log("Application template has been stamped.");
  	        // Now that the template has been stamped, you can use 'automatic node finding' $ syntax to access widgets.
  	        // e.g. to access a widget with an id of 'widget_id' you can use templateObj.$.widgetId

  	        templateObj.$.btn_connect.onclick = wss_connect;
            templateObj.$.btn_connect.disabled = false;
            templateObj.$.chk_autortls.checked = true;
            templateObj.$.chk_autortls.disabled = false;

            templateObj.$.btn_scan.onclick = scan_for_slave;
            templateObj.$.btn_scan.disabled = true;
            templateObj.$.btn_conn_disconn.disabled = true;
            templateObj.$.btn_conn_disconn.onclick = connect_disconnect;
            templateObj.$.btn_enable_rtls.disabled = true;
            templateObj.$.btn_enable_rtls.onclick = enable_rtls;

            templateObj.$.ti_widget_progressbar.visible = false;
            templateObj.$.ti_widget_status_label.visible = false;

            templateObj.$.btn_reload.onclick = () => document.location.reload();

        },1);

	});
};

templateObj = document.querySelector('#template_obj');
if (templateObj) {
    init();
} else {
    document.addEventListener('DOMContentLoaded',init.bind(this))
}


